{ Kalman filter for 3-state , 1 -measurement, steady solution, with hit rejection

'F'-means takes float values as arguments
Units:
*time (ticks) - INT internal & external  , no conversion INT<-->FLOAT
*stat vector X - 100*cm, internal=FLOAT, I/O=INT <-- needs conversion  

}

CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  _1sec=80_000_000 ' 1 second
  f_1sec=float(_1sec)
  f_1sec2=float(_1sec)*float(_1sec) 
  _1msec=_1sec/1000 ' mili sec

  cX2atten=0.98 ' speed attenuation if y1 rejected
  cX3atten=0.1 ' acceletarion attenuation if y1 rejected  
VAR
  ' float values
  long  mfX1,mfX2,mfX3 ' current state   
  long  mfX1pre,mfX2pre,mfX3pre ' prediction (apriori)
  long  cfK1,cfK2,cfK3 ' constant kalman gains
  long  cfA0,cfA1,cfA2,cfB0,cfB1,cfB2
  long  cfNsigY ' scale for sigX1 for cut-off
  long   mfw1,mfw2 ' working variables for QA
  ' integer values
  long  miT, miTpre   ' time in ticks
  long  miSigX1s ' predicted error of X1, scaled by cNsigY
  long  ciDt1,ciDt2  ' time ranges in ticks
  long  ciMaxY ' limits on Y1
  long  ciX1cent,ciX1marg ' braket range of X1
OBJ
  f:"Float32"
  'f:"DynamicMathLib"    
PUB start
    f.start
PUB setState(x1,x2,x3,t0cnt)
  mfX1:=x1
  mfX2:=x2
  mfX3:=x3
  miT:=t0cnt
  
PUB  setKgains(fk1,fk2,fk3)
  cfK1:=fk1
  cfK2:=fk2
  cfK3:=fk3

PUB  setCutLimitsI(ic1,ic2)
  ciMaxY:=ic1
  cfNsigY:=f.ffloat(ic2)

PUB setX1bracket(xc,xm)
  ciX1cent:=xc
  ciX1marg:=xm
pub getT ' unit= ticks 
    return miT
       
pub getX1 ' unit= 100*cm
       return f.fround(mfX1)
       
pub getW1 
       return f.fround(mfW1)
pub getW2 
       return f.fround(mfW2)
       
pub getSigX1s 
    return miSigX1s
       
pub getX2sec ' unit= 100*cm/clockTick  *1sec   
    return f.fround(f.FMul(mfX2,f_1sec))
       
pub getX3sec2 ' unit= 100*cm/clockTick^2 *1sec^2
    return f.fround(f.FMul(mfX3,f_1sec2))
       
PUB setX1error(dt1,dt2,a0,b0,a1,b1,a2,b2)
   ciDt1:=dt1 ' clockTick
   ciDt2:=dt2 ' clockTick
   cfA0:=a0   '100*cm
   cfA1:=a1   '100*cm  
   cfA2:=a2   '100*cm  
   cfB0:=b0   '100*cm/clockTick
   cfB1:=b1   '100*cm/clockTick
   cfB2:=f.FMul(b2,2.0)   '100*cm/clockTick&2  , extra 2 to match use below

PUB extrapolateI(t1cnt) |idt,fdt,fdt2,a,b,sigX1
   miTpre:=t1cnt
   idt:=miTpre-miT ' time diff in ticks  
   fdt:= f.FFloat(idt)
   fdt2:=f.FMul(fdt,fdt)
   fdt2:=f.FMul(fdt2,0.5)  '=Dt*Dt/2
   'return dt
   '.... compute apriori state vector components
   '----X1----
    a:=f.FMul(mfX2,fdt)
    b:=f.FMul(mfX3,fdt2)
    mfX1pre:=f.FAdd(mfX1,a)
    mfX1pre:=f.FAdd(mfX1pre,b)
   
  '----X2----
    a:=f.FMul(mfX3,fdt)
    mfX2pre:=f.FAdd(mfX2,a)

  '----X3----
   mfX3pre:=mfX3

   mfW1:=mfX1pre 'tmp
  
  '----- sigX1pre --- 
  if idt >ciDt2
      a:=f.FMul(cfB2,fdt2)
        sigX1:=f.FAdd(cfA2,a)
  elseif idt > ciDt1
       a:=f.FMul(cfB1,fdt)
        sigX1:=f.FAdd(cfA1,a)
  else      
       a:=f.FMul(cfB0,fdt)
        sigX1:=f.FAdd(cfA0,a)
             
  miSigX1s:=f.fround(f.fmul(sigX1,cfNsigY))

 
PUB correctI(iy1) | fdy,a ,iX1pre ,err
   'iy1=INT, 100*cm - custom X-units
   
   iX1pre:=f.Fround(mfX1pre)
   err:=0
   if ciMaxY <iy1
      err+=1

   if||(iy1-iX1pre)> miSigX1s
      err+=2
      
   if||(ciX1cent-iX1pre)> ciX1marg
      err+=4
      mfX1pre:=f.ffloat(ciX1cent)
      mfX2pre:=mfX3pre:=0.0
      miT:=miTpre
        
   if err>0 '...only extrapolation was used
      mfX1:=mfX1pre
      mfX2:=f.fmul(mfX2pre,cX2atten)
      mfX3:=f.fmul(mfX3pre,cX3atten)  
   else
                       
   '... incorporate innovation
     a:=f.ffloat(iy1) 
     fdy:=f.fSub(a,mfX1pre) ' innovation, in 100*cm   
     a:=f.FMul(cfK1,fdy)
     mfX1:=f.FAdd(mfX1pre,a)
     a:=f.FMul(cfK2,fdy)
     mfX2:=f.FAdd(mfX2pre,a)
     a:=f.FMul(cfK3,fdy)
     mfX3:=f.FAdd(mfX3pre,a)
     miT:=miTpre
   return err                         